﻿//#define USE_DUNGEONBUDDY_DLL
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Frames;
using Styx.Helpers;
using Styx.Pathing;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Action = Styx.TreeSharp.Action;
#if USE_DUNGEONBUDDY_DLL
using Bots.DungeonBuddyDll;
using Bots.DungeonBuddyDll.Profiles;
using Bots.DungeonBuddyDll.Attributes;
using Bots.DungeonBuddyDll.Helpers;

namespace Bots.DungeonBuddyDll.Dungeon_Scripts.Classic
#else
    using Bots.DungeonBuddy.Profiles;
    using Bots.DungeonBuddy.Attributes;
    using Bots.DungeonBuddy.Helpers;
namespace Bots.DungeonBuddy.Dungeon_Scripts.Mists_of_Pandaria
#endif

{
    public class Scholomance : Dungeon
    {
        #region Overrides of Dungeon

        public override uint DungeonId
        {
            get { return 2; }
        }

        public override WoWPoint Entrance
        {
            get { return new WoWPoint(1279.48, -2551.52, 87.41); }
        }

        public override WoWPoint ExitLocation
        {
            get { return new WoWPoint(188.05, 126.49, 138.82); }
        }

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            units.RemoveAll(
                ret =>
                {
                    var unit = ret as WoWUnit;
                    if (unit != null)
                    {
                        if (unit.Entry == JandiceBarovSpawnId && unit.DisplayId != JandiceDisplayId)
                            return true;

                        if (unit.Entry == FailedStudentId && Me.IsDps() && Me.IsMelee() && unit.HealthPercent < 30)
                            return true;

                        if (unit.Entry == WovenBoneguardId && unit.Combat && Me.Combat && !unit.IsTargetingMeOrPet && !unit.IsTargetingMyPartyMember)
                            return true;

                        if (unit.Entry == LiliansSoulId && Me.IsMelee() && _isFixated)
                            return true;

                        if ((unit.Entry == ExpiredTestSubjectId || unit.Entry == FailedStudentId) && unit.CurrentTargetGuid > 0)
                        {
                            var expiredTestSubjectTarget = unit.CurrentTarget as WoWPlayer;
                            if (expiredTestSubjectTarget != null && !expiredTestSubjectTarget.IsMe && IsTakingHarshLesson(expiredTestSubjectTarget))
                                return true;
                        }
                        if (unit.Entry == DarkmasterGandlingId && IsTakingHarshLesson(Me))
                            return true;
                        // ignore adds if they're not near boss as melee.. everyone should group up on tank when not taking harsh lessons to aoe down adds and aoe heals.
                        if (_darkmasterGandling != null && _darkmasterGandling.IsValid && _darkmasterGandling.Combat && unit != _darkmasterGandling &&
                            !ScriptHelpers.IsBossAlive("Lilian Voss") && Me.IsMelee() && unit.Location.Distance(_darkmasterGandling.Location) > 15)
                            return true;
                    }
                    return false;
                });
        }

        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {
            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit != null)
                {
                    if (unit.Entry == BoneweaverId && Me.IsTank() && Me.Combat && !ScriptHelpers.IsBossAlive("Jandice Barov"))
                        outgoingunits.Add(unit);

                    if (unit.Entry == InstructorChillheartsPhylactery)
                        outgoingunits.Add(unit);

                    if (unit.Entry == JandiceBarovSpawnId && unit.DisplayId == JandiceDisplayId)
                        outgoingunits.Add(unit);

                    if (unit.Entry == ExpiredTestSubjectId && IsTakingHarshLesson(Me))
                        outgoingunits.Add(unit);
                }
            }
        }

        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit != null)
                {
                    if (unit.Entry == BoneweaverId)
                        priority.Score += 1000;

                    if (unit.Entry == CandlestickMageId && StyxWoW.Me.IsDps())
                        priority.Score += 500;

                    if (unit.Entry == MeatGraftId && StyxWoW.Me.IsDps())
                        priority.Score += 5000;

                    if (unit.Entry == FailedStudentId && !IsTakingHarshLesson(Me))
                        priority.Score += 5000;
                }
            }
        }

        #endregion

        private const uint InstructorChillheartsPhylactery = 58664;
        private const uint CandlestickMageId = 59467;
        private const uint JandiceBarovSpawnId = 59220;
        private const uint BoneweaverId = 59193;
        private const uint BonePileId = 59304;
        private const uint SoulflameId = 59316;
        private const uint MeatGraftId = 59982;
        private const uint LiliansSoulId = 58791;
        private const uint WovenBoneguardId = 59213;
        private const uint LilianVossId = 58722;
        private const uint ExpiredTestSubjectId = 59100;
        private const uint FreshTestSubjectId = 59099;
        private const uint BoredStudentId = 59614;
        private const uint DarkmasterGandlingId = 59080;
        private const uint FailedStudentId = 59078;
        private const int JandiceDisplayId = 43460;
        private const uint RattlegoreId = 59153;
        private readonly Dictionary<ulong, WaitTimer> _harshLessons = new Dictionary<ulong, WaitTimer>();
        private readonly WaitTimer _updateHarshLessonTimer = new WaitTimer(TimeSpan.FromSeconds(1));
        private WoWUnit _darkmasterGandling;
        private bool _isFixated;

        private static LocalPlayer Me
        {
            get { return StyxWoW.Me; }
        }

        [EncounterHandler(0, "Root")]
        public Composite RootEncounter()
        {
            //  Me.CurrentTarget.DisplayId
            return new PrioritySelector();
        }

        [EncounterHandler(64563, "Talking Skull", Mode = CallBehaviorMode.Proximity)]
        [EncounterHandler(64562, "Talking Skull", Mode = CallBehaviorMode.Proximity)]
        public Composite TalkingSkullEncounter()
        {
            WoWUnit unit = null;
            return new PrioritySelector(
                ctx => unit = ctx as WoWUnit,
                new Decorator(ctx => unit.QuestGiverStatus == QuestGiverStatus.Available, ScriptHelpers.CreatePickupQuest(() => unit)),
                new Decorator(ctx => unit.QuestGiverStatus == QuestGiverStatus.TurnIn, ScriptHelpers.CreateTurninQuest(() => unit)));
        }


        // pickup quest items.
        [ObjectHandler(214101, "In the Shadow of the Light", ObjectRange = 30)]
        [ObjectHandler(214279, "In the Shadow of the Light Heroic", ObjectRange = 30)]
        [ObjectHandler(214105, "Kel'Thuzad's Deep Knowledge", ObjectRange = 30)]
        [ObjectHandler(214278, "Kel'Thuzad's Deep Knowledge Heroic", ObjectRange = 30)]
        [ObjectHandler(214106, "Forbidden Rites and other Rituals Necromantic", ObjectRange = 30)]
        [ObjectHandler(214280, "Forbidden Rites and other Rituals Necromantic Heroic", ObjectRange = 30)]
        [ObjectHandler(214107, "The Dark Grimoire", ObjectRange = 30)]
        [ObjectHandler(214277, "The Dark Grimoire Heroic", ObjectRange = 30)]
        public Composite TheFourTombsHandler()
        {
            WoWGameObject questObject = null;
            var heroicQuestObjectiveIndex = new Dictionary<uint, int>
            {
                {214279, 0}, // In the Shadow of the Light
                {214278, 1}, // Kel'Thuzad's Deep Knowledge
                {214280, 2}, // Forbidden Rites and other Rituals Necromantic 
                {214277, 3} // The Dark Grimoire
            };
            var normalQuestObjectiveIndex = new Dictionary<uint, int>
            {
                {214101, 0}, // In the Shadow of the Light
                {214105, 1}, // Kel'Thuzad's Deep Knowledge
                {214106, 2}, // Forbidden Rites and other Rituals Necromantic 
                {214107, 3} // The Dark Grimoire
            };
            var theFourTomesIds = new uint[] { 31442, 31440 };
            WoWDescriptorQuest questDesc;

            return new PrioritySelector(
                ctx => questObject = ctx as WoWGameObject,
                new Decorator(
                    ctx =>
                    {
                        var quest = Me.QuestLog.GetAllQuests().FirstOrDefault(q => theFourTomesIds.Contains(q.Id));
                        return quest != null && !quest.IsCompleted && quest.GetData(out questDesc) &&
                               questDesc.ObjectivesDone[(DungeonId == 2 ? normalQuestObjectiveIndex : heroicQuestObjectiveIndex)[questObject.Entry]] == 0 &&
                               Targeting.Instance.IsEmpty() && !ScriptHelpers.WillPullAggroAtLocation(questObject.Location) &&
                               Me.Location.PathDistance(questObject.Location) <= 30;
                    },
                    new PrioritySelector(
                        new Decorator(ctx => !questObject.WithinInteractRange, new Action(ctx => Navigator.MoveTo(questObject.Location))),
                        new Decorator(
                            ctx => questObject.WithinInteractRange,
                            new Sequence(
                                new Action(ctx => questObject.Interact()),
                                new WaitContinue(3, ctx => LootFrame.Instance.IsVisible, new ActionAlwaysSucceed()),
                                new Action(ctx => Logger.Write("Looting {0}", questObject.Name)),
                                new Action(ctx => LootFrame.Instance.LootAll()))))));
        }

        [EncounterHandler(58633, "Instructor Chillheart", Mode = CallBehaviorMode.Proximity, BossRange = 80)]
        public Composite InstructorChillheartEncounter()
        {
            const uint iceWallId = 62731;
            const uint arcaneBombId = 58753;
            const uint antonidasSelfHelpGuideToStandingInFireId = 58635;
            const uint frigidGraspId = 58640;

            var phase1TankLoc = new WoWPoint(203.2554, 38.66046, 119.2258);
            const uint risenGuard = 58822;

            WoWUnit boss = null;
            AddAvoidObject(ctx => true, 5, frigidGraspId);
            AddAvoidObject(ctx => true, 3, u => u is WoWPlayer && !u.IsMe && ((WoWPlayer)u).HasAura("Ice Wrath"));
            AddAvoidObject(ctx => true, 4, arcaneBombId, antonidasSelfHelpGuideToStandingInFireId);
            // avoid the icewall. 
            AddAvoidObject(
                ctx => true,
                8,
                u => u.Entry == iceWallId,
                o =>
                { // the icewall object is just one NPC so we need to use its y location and use our x location 
                    var loc = o.Location;
                    loc.X = Me.X;
                    return loc;
                });
            return new PrioritySelector(
                ctx => boss = ctx as WoWUnit,
                // OOC behavior
                new Decorator(
                    ctx => !boss.Combat,
                    new PrioritySelector(
                // kill all risen guards 
                        ScriptHelpers.CreateClearArea(() => boss.Location, 70, u => u.Entry == risenGuard))),
                // Phase 1 combat behavior
                new Decorator(
                    ctx => boss.Combat && !boss.HasAura("Permanent Feign Death"),
                    new PrioritySelector(
                // move to the west wall 
                        new Decorator(
                            ctx => StyxWoW.Me.IsRange() && StyxWoW.Me.Y < 30,
                            new Action(
                                ctx =>
                                {
                                    var loc = StyxWoW.Me.Location;
                                    loc.Y = 35;
                                    Navigator.MoveTo(loc);
                                })),
                        ScriptHelpers.CreateTankUnitAtLocation(() => phase1TankLoc, 5))),
                // Phase 2 combat behavior
                new Decorator(ctx => boss.Combat && boss.HasAura("Permanent Feign Death"), new PrioritySelector()));
        }

        [EncounterHandler(59184, "Jandice Barov", Mode = CallBehaviorMode.CurrentBoss)]
        public Composite JandiceBarovIllusionEncounter()
        {
            var combatTimer = new WaitTimer(TimeSpan.FromSeconds(6));

            WoWUnit boss = null;
            return new PrioritySelector(
                ctx => boss = ctx as WoWUnit,
                // handle vanish.
                new Decorator(
                    ctx => boss != null && boss.Combat,
                    new Action(
                        ctx =>
                        {
                            combatTimer.Reset();
                            return RunStatus.Failure;
                        })),
                new Decorator(ctx => boss == null && !combatTimer.IsFinished && Targeting.Instance.IsEmpty() && StyxWoW.Me.IsTank(), new ActionAlwaysSucceed()));
        }

        [EncounterHandler(59184, "Jandice Barov")]
        public Composite JandiceBarovEncounter()
        {
            const int wondrousRapidityId = 114062;
            const uint gravityFluxId = 114035;
            WoWUnit boss = null;
            AddAvoidObject(ctx => true, 8, o => o is WoWAreaTrigger && ((WoWAreaTrigger)o).SpellId == gravityFluxId);
            return new PrioritySelector(
                ctx => boss = ctx as WoWUnit,
                ScriptHelpers.CreateAvoidUnitAnglesBehavior(
                    ctx => Me.IsTank() && boss.CastingSpellId == wondrousRapidityId || !Me.IsTank(), () => boss, new ScriptHelpers.AngleSpan(0, 90)));
        }

        [EncounterHandler(59153, "Rattlegore", BossRange = 100, Mode = CallBehaviorMode.Proximity)]
        public Composite RattlegoreEncounter()
        {
            WoWUnit boss = null;
            WoWUnit selectedBonePile = null;
            var waitLoc = new WoWPoint(261.0993, 64.81649, 113.4927);
            var tankLoc = new WoWPoint(261.4132, 99.62674, 113.4891);
            var roomCenter = new WoWPoint(263.8696, 99.24605, 113.489);
            // run from soul flame.
            AddAvoidObject(ctx => true, obj => obj.Scale, SoulflameId);
            AddAvoidObject(ctx => Me.IsTank(), 10, o => o.Entry == RattlegoreId);
            return new PrioritySelector(
                ctx =>
                {
                    var soulFlames = ObjectManager.GetObjectsOfType<WoWUnit>().Where(u => u.Entry == SoulflameId).ToArray();
                    selectedBonePile = (from bonePile in ObjectManager.GetObjectsOfType<WoWUnit>().Where(u => u.Entry == BonePileId)
                                        let bonepileDist = bonePile.Distance
                                        // skip any bone piles that are inside a soul flame.
                                        where
                                            soulFlames.All(s => s.Location.Distance(bonePile.Location) > s.Scale + 2) &&
                                            // skip bone piles that other party members who are missing the buff, are heading towards it and are closer.
                                            !Me.PartyMembers.Any(
                                                p =>
                                                !p.HasAura("Bone Armor") && p.IsSafelyFacing(bonePile, 45) && p.IsMoving &&
                                                p.Location.Distance(bonePile.Location) < bonepileDist)
                                        orderby bonePile.Distance2DSqr
                                        select bonePile).FirstOrDefault();
                    selectedBonePile =
                        ObjectManager.GetObjectsOfType<WoWUnit>().Where(u => u.Entry == BonePileId && soulFlames.All(s => s.Location.Distance(u.Location) > s.Scale + 2)).
                            OrderBy(u => u.DistanceSqr).FirstOrDefault();
                    return boss = ctx as WoWUnit;
                },
                // move inside the gate.
                new Decorator(ctx => !Me.IsTank() && ScriptHelpers.Tank != null && ScriptHelpers.Tank.Y >= 61 && Me.Y < 61f, new Action(ctx => Navigator.MoveTo(waitLoc))),
                new Decorator(
                    ctx => Me.IsTank() && !Me.Combat && Me.Y >= 61 && Me.PartyMembers.Count(p => p.Y >= 61f) < 4,
                    new Sequence(new ActionSetActivity("Waiting for other party members to enter room at Rattlegore"), new Action(ctx => Navigator.MoveTo(waitLoc)))),
                // get the bone pile buff.
                new Decorator(
                    ctx => boss.Combat && !Me.HasAura("Bone Armor") && selectedBonePile != null && (Me.IsTank() && boss.CurrentTargetGuid == Me.Guid || !Me.IsTank()),
                    new PrioritySelector(
                        new Decorator(ctx => selectedBonePile.WithinInteractRange, new Action(ctx => selectedBonePile.Interact())),
                        new Decorator(ctx => !selectedBonePile.WithinInteractRange, new Action(ctx => Navigator.MoveTo(selectedBonePile.Location))))),
                new Decorator(
                    ctx =>
                    Me.IsTank() && boss.CurrentTargetGuid == Me.Guid && Me.HasAura("Bone Armor") &&
                    (boss.HasAura("Rusting") && boss.Auras["Rusting"].StackCount <= 3 || !boss.HasAura("Rusting")),
                    ScriptHelpers.CreateTankUnitAtLocation(() => tankLoc, 5)));
        }

        [EncounterHandler(58722, "Lilian Voss", Mode = CallBehaviorMode.Proximity)]
        public Composite LilianVossEncounter()
        {
            const uint darkBlazeId = 58780;
            const int deathsGraspId = 111570;
            const int shadowShivId = 111775;
            var runFromBossTimer = new WaitTimer(TimeSpan.FromSeconds(3));
            var entLoc = new WoWPoint(200.3293, 104.8955, 108.2328);
            WoWUnit boss = null;
            var roomCenterLoc = new WoWPoint(200.4618, 86.05209, 107.7617);
            // prevent bot from going into hallway during encounter.
            AddAvoidLocation(ctx => boss != null && boss.IsValid && boss.Combat && Me.Y <= 104, 4, o => entLoc);
            // run from from the fire on ground. ignore it if moving while fixated.
            AddAvoidObject(ctx => _isFixated && !Me.IsMoving || !_isFixated, () => roomCenterLoc, 25, 2, darkBlazeId);
            // run away from boss after she casts deaths grasp.
            AddAvoidObject(ctx => boss.CastingSpellId != deathsGraspId && !runFromBossTimer.IsFinished, 8, u => u == boss);
            // run from shadow shiv target
            AddAvoidObject(
                ctx => true,
                8,
                u =>
                u is WoWPlayer && (boss != null && boss.IsValid && boss.CastingSpellId == shadowShivId && boss.CurrentTargetGuid == u.Guid || u.ToUnit().HasAura(shadowShivId)));
            return new PrioritySelector(
                ctx =>
                {
                    boss = ctx as WoWUnit;
                    if (boss.CastingSpellId == deathsGraspId)
                        runFromBossTimer.Reset();
                    return boss;
                },
                new Decorator(
                    ctx => boss.Combat && ScriptHelpers.IsBossAlive("Lilian Voss"),
                    new PrioritySelector(
                        new Decorator(ctx => Me.Y > 104, new Action(ctx => Navigator.PlayerMover.MoveTowards(roomCenterLoc))),
                        new Decorator(ctx => Me.IsTank() && boss.IsTargetingMeOrPet, ScriptHelpers.CreateTankUnitAtLocation(() => roomCenterLoc, 20)))));
        }

        [EncounterHandler(58791, "Lilian's Soul")]
        public Composite LiliansSoulEncounter()
        {
            const int fixateAngerId = 115350;
            WoWUnit boss = null;
            var roomCenterLoc = new WoWPoint(200.4618, 86.05209, 107.7617);

            // run from bot if fixated. _isFixated is updated in the behavior.
            AddAvoidObject(ctx => _isFixated, () => roomCenterLoc, 25, 10, LiliansSoulId);
            return new PrioritySelector(
                ctx =>
                {
                    boss = ctx as WoWUnit;
                    _isFixated = boss.CastingSpellId == fixateAngerId && boss.CurrentTargetGuid == Me.Guid || Me.HasAura(115350);
                    return boss;
                },
                // stay out of the hall to avoid getting cornered.
                new Decorator(ctx => Me.Y > 104, new Action(ctx => Navigator.PlayerMover.MoveTowards(roomCenterLoc))),
                new Decorator(ctx => Me.IsTank() && Targeting.Instance.IsEmpty(), new ActionAlwaysSucceed()));
        }

        [ObjectHandler(211277, "Coffer of Forgotten Souls")]
        [ObjectHandler(211278, "Coffer of Forgotten Souls")]
        public Composite CofferofForgottenSoulsHandler()
        {
            WoWGameObject chest = null;
            return new PrioritySelector(
                ctx => chest = ctx as WoWGameObject,
                new Decorator(
                    ctx => chest.CanLoot,
                    new PrioritySelector(
                        new Decorator(ctx => ScriptHelpers.IsBossAlive("Lilian Voss"), new Action(ctx => ScriptHelpers.MarkBossAsDead("Lilian Voss"))),
                        ScriptHelpers.CreateInteractWithObject(() => chest))));
        }


        [EncounterHandler(59613, "Professor Slate", Mode = CallBehaviorMode.Proximity, BossRange = 62)]
        public Composite ProfessorSlateEncounter()
        {
            var boredStudenPullToLoc = new WoWPoint(144.916, 73.33847, 106.3931);

            WoWUnit boss = null, boredStudent = null;
            const uint toxicPoison = 114873;
            AddAvoidObject(ctx => true, 4, toxicPoison);

            return new PrioritySelector(
                ctx => boss = ctx as WoWUnit,
                new Decorator(
                    ctx => !boss.Combat && Me.Z < 110,
                    new PrioritySelector(
                        ctx =>
                        boredStudent =
                        ObjectManager.GetObjectsOfType<WoWUnit>().Where(u => u.Entry == BoredStudentId && u.IsAlive && u.Z < 106).OrderBy(
                            u => u.Location.PathDistance(Me.Location)).FirstOrDefault(),
                        ScriptHelpers.CreatePullNpcToLocation(ctx => boredStudent != null, () => boredStudent, () => boredStudenPullToLoc, 6))),
                new Decorator(
                    ctx => boss.Combat,
                    new PrioritySelector(
                        ScriptHelpers.CreateTankFaceAwayGroupUnit(() => boss, 15),
                        ScriptHelpers.CreateAvoidUnitAnglesBehavior(ctx => !Me.IsTank() && boss.Distance < 15, () => boss, new ScriptHelpers.AngleSpan(0, 60)))));
        }

        #region Darkmaster Gandling

        [EncounterHandler(59080, "Darkmaster Gandling", BossRange = 300)]
        public Composite DarkmasterGandlingEncounter()
        {
            // avoid running into doors when having a harse lesson.
            AddAvoidObject(
                ctx => _darkmasterGandling != null && _darkmasterGandling.IsValid && _darkmasterGandling.Combat && IsTakingHarshLesson(Me),
                4,
                u => u is WoWGameObject && u.ToGameObject().SubType == WoWGameObjectType.Door && u.ZDiff < 10);
            WoWSpell dispell = null;
            WoWUnit dispellTarget = null;

            WoWUnit[] freshTestSubjects = null;

            return new PrioritySelector(
                ctx => _darkmasterGandling = ctx as WoWUnit,
                new ActionSetActivity("I am {0} taking a harsh lesson", IsTakingHarshLesson(Me) ? "" : "not"),
                ScriptHelpers.CreateDispellParty("Immolate", ScriptHelpers.PartyDispellType.Magic),
                // move to tank if not taking harsh lesson
                new PrioritySelector(
                    ctx =>
                    {
                        var tank = ScriptHelpers.Tank;
                        return !IsTakingHarshLesson(tank) && tank.IsAlive ? tank.Location : _darkmasterGandling.Location;
                    },
                new Decorator<WoWPoint>(
                    loc => !Me.IsTank() && !IsTakingHarshLesson(Me) && loc.Distance(Me.Location) > 6,
                    new Helpers.Action<WoWPoint>(loc => Navigator.MoveTo(loc)))),
                new Decorator(
                // healer behavior for harsh lesson.
                    ctx => IsTakingHarshLesson(Me) && Me.IsHealer(),
                    new PrioritySelector(
                        ctx =>
                        {
                            freshTestSubjects =
                                ObjectManager.GetObjectsOfType<WoWUnit>().Where(
                                    u => u.Entry == FreshTestSubjectId && u.HasAura("Explosive Pain") && Me.Location.PathDistance(u.Location) < 50).OrderBy(
                                        u => u.DistanceSqr).ToArray();

                            dispellTarget =
                                freshTestSubjects.FirstOrDefault(
                                    u => ObjectManager.GetObjectsOfType<WoWUnit>().Count(o => o.Entry == ExpiredTestSubjectId && o.Location.Distance(u.Location) < 8) >= 1);

                            dispell = ScriptHelpers.GetDefensiveMagicDispell();
                            return dispellTarget;
                        },
                        new Decorator(
                            ctx => dispellTarget != null && dispell != null && SpellManager.CanCast(dispell, dispellTarget, true, false),
                            new Action(ctx => SpellManager.Cast(dispell, dispellTarget))),
                        new Decorator(
                            ctx => dispellTarget != null && dispell != null && (dispellTarget.Distance > 30 || !dispellTarget.InLineOfSpellSight),
                            new Action(ctx => Navigator.MoveTo(dispellTarget.Location))),
                        new Decorator(
                            ctx => dispellTarget == null && dispell != null && freshTestSubjects.Length > 0 && freshTestSubjects[0].Distance > 4,
                            new Action(ctx => Navigator.MoveTo(freshTestSubjects[0].Location))))));
        }

        private bool IsTakingHarshLesson(WoWPlayer player)
        {
            if (_updateHarshLessonTimer.IsFinished)
            {
                var raidGuids = Me.GroupInfo.RaidMemberGuids;
                var keys = _harshLessons.Keys.ToList();
                foreach (var guid in keys)
                {
                    if (!raidGuids.Contains(guid))
                    {
                        _harshLessons.Remove(guid);
                        Logger.Write("{0} will nolonger receive harsh lessons", guid);
                    }
                }

                var raidMembers = Me.RaidMembers;
                foreach (var guid in raidGuids)
                {
                    var p = raidMembers.FirstOrDefault(r => r.Guid == guid);
                    if (p == null) continue;
                    bool hasHarshLessonAura = p.HasAura("Harsh Lesson");
                    if (!hasHarshLessonAura && _harshLessons.ContainsKey(guid))
                    {
                        _harshLessons.Remove(guid);
                        Logger.Write("{0} doesn't have harsh lesson aura.", p.UnitId());
                    }
                    else if (hasHarshLessonAura && !_harshLessons.ContainsKey(guid))
                    {
                        Logger.Write("{0} just started taking harsh lessons", p.UnitId());
                        _harshLessons.Add(guid, new WaitTimer(TimeSpan.FromSeconds(28)));
                        _harshLessons[guid].Reset();
                    }
                    if (p.IsMe && hasHarshLessonAura && !_harshLessons[Me.Guid].IsFinished)
                        Logger.Write("I am taking a harsh lesson.");
                }
                _updateHarshLessonTimer.Reset();
            }
            var playerGuid = player.Guid;
            var returnVal = _harshLessons.ContainsKey(playerGuid) && !_harshLessons[playerGuid].IsFinished;

            return returnVal;
        }

        #endregion

    }
}