﻿//#define USE_DUNGEONBUDDY_DLL
using System;
using System.Collections.Generic;
using System.Linq;
using CommonBehaviors.Actions;
using MainDev.RemoteASM.Handlers;
using Styx;
using Styx.Common;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Frames;
using Styx.Helpers;
using Styx.Patchables;
using Styx.Pathing;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Styx.WoWInternals.World;
using Action = Styx.TreeSharp.Action;

#if USE_DUNGEONBUDDY_DLL
using Bots.DungeonBuddyDll;
using Bots.DungeonBuddyDll.Profiles;
using Bots.DungeonBuddyDll.Attributes;
using Bots.DungeonBuddyDll.Helpers;
using AvoidanceManager = Bots.DungeonBuddyDll.Avoidance.AvoidanceManager;

namespace Bots.DungeonBuddyDll.Dungeon_Scripts.Mists_of_Pandaria
#else
    using Bots.DungeonBuddy.Profiles;
    using Bots.DungeonBuddy.Attributes;
    using Bots.DungeonBuddy.Helpers;
    using AvoidanceManager = Bots.DungeonBuddy.Avoidance.AvoidanceManager;
namespace Bots.DungeonBuddy.Dungeon_Scripts.Mists_of_Pandaria
#endif
{
    public class TheVaultOfMysteries : Dungeon
    {
        #region Overrides of Dungeon

        private const float ElegonInnerCircleRadius = 46;
        private const float ElegonOuterCircleRadius = 63;
        private readonly WoWPoint _elegonRoomCenterLoc = new WoWPoint(4023.042, 1907.786, 358.7884); // inner radius 46. outer radius 63
        private readonly WoWPoint _elegonRoomEntranceLoc = new WoWPoint(4035.819, 1861.859, 358.7884);

        public override uint DungeonId
        {
            get { return 528; }
        }

        public override WoWPoint Entrance
        {
            get { return new WoWPoint(3978.82, 1115.574, 497.136); }
        }

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            units.RemoveAll(
                ret =>
                {
                    var unit = ret.ToUnit();
                    if (unit != null)
                    {
                        if (unit.Entry == ElegonId && unit.HasAura(""))
                            return true;
                    }
                    return false;
                });
        }

        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {
            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit != null)
                {
                    if (unit.Entry == PinningArrowId && Me.IsRange())
                        outgoingunits.Add(unit);

                    else if (unit is WoWPlayer && unit.HasAura("Maddening Shout"))
                        outgoingunits.Add(unit);

                    else if (unit.Entry == UndyingShadowsId && Me.IsRange())
                        outgoingunits.Add(unit);

                    else if (unit.Entry == EmpyrealFocusId && unit.CanSelect)
                        outgoingunits.Add(unit);
                }
            }
        }

        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit != null)
                {
                    if (unit.Entry == PinningArrowId && Me.IsRange())
                        priority.Score += 1000;

                    else if (unit.Entry == UndyingShadowsId && Me.IsRange())
                        priority.Score += 1000;

                    else if (unit is WoWPlayer)
                        priority.Score += 1000;

                    else if (unit.Entry == EnergyChargeId)
                        priority.Score += 1000;

                    else if (unit.Entry == EmpyrealFocusId)
                        priority.Score += 1000;

                    else if (unit.Entry == CelestialProtectorId)
                        priority.Score += 1000;

                    else if (unit.Entry == EmperorsCourageId && (Me.IsMelee() && unit.Distance <= 50 || !Me.IsMelee()))
                        priority.Score += 50000;

                    else if (_emperorsStrengthIds.Contains(unit.Entry) && (Me.IsMelee() && unit.Distance <= 50 || !Me.IsMelee()))
                        priority.Score += 4000;

                    else if (unit.Entry == EmperorsRageId && (Me.IsMelee() && unit.Distance <= 50 || !Me.IsMelee()))
                        priority.Score += 3000;
                    else if (unit.Entry == QinxiId || unit.Entry == JanxiId)
                        priority.Score = 800 - unit.Distance;
                }
            }
        }

        public override void OnEnter()
        {
            if (Me.IsTank())
            {
                Alert.Show(
                    "Tanking Not Supported",
                    string.Format(
                        "Tanking is not supported in the {0} script. If you wish to stay in raid and play manually then press 'Continue'. Otherwise you will automatically leave raid.",
                        Name),
                    30,
                    true,
                    true,
                    null,
                    () => Lua.DoString("LeaveParty()"),
                    "Continue",
                    "Leave");
            }
            else
            {
                Alert.Show(
                    "Do Not AFK",
                    "It is highly recommended you do not afk while in a raid and be prepared to intervene if needed in the event something goes wrong or you're asked to perform a certain task.",
                    20,
                    true,
                    false,
                    null,
                    null,
                    "Ok");
            }
        }

        public override MoveResult MoveTo(WoWPoint location)
        {
            var myLoc = Me.Location;
            var destDistFromElegonCenter = location.Distance(_elegonRoomCenterLoc);
            var myDistFromElegonCenter = myLoc.Distance(_elegonRoomCenterLoc);
            // check if we want to move onto the inner circle area.
            if (destDistFromElegonCenter <= ElegonInnerCircleRadius)
            {
                if (myDistFromElegonCenter > ElegonOuterCircleRadius)
                    return Navigator.MoveTo(_elegonRoomEntranceLoc);
                if (myDistFromElegonCenter <= ElegonInnerCircleRadius + 2)
                {
                    Navigator.PlayerMover.MoveTowards(location);
                    return MoveResult.Moved;
                }
                var moveTo = GetElegonInnerCircleMovetoPoints(ElegonInnerCircleRadius).OrderBy(l => l.DistanceSqr(myLoc)).FirstOrDefault();
                return Navigator.MoveTo(moveTo);
            }
            if (myDistFromElegonCenter <= ElegonInnerCircleRadius)
            {
                var moveTo = GetElegonInnerCircleMovetoPoints(ElegonInnerCircleRadius + 2).OrderBy(l => l.DistanceSqr(location)).FirstOrDefault();
                Navigator.PlayerMover.MoveTowards(moveTo);
                return MoveResult.Moved;
            }
            return base.MoveTo(location);
        }

        private IEnumerable<WoWPoint> GetElegonInnerCircleMovetoPoints(float dist)
        {
            const float pix2 = (float)(Math.PI * 2);
            for (float ang = 0; ang < 360; ang += 40)
            {
                var radians = WoWMathHelper.DegreesToRadians(ang);
                yield return _elegonRoomCenterLoc.RayCast(radians, dist);
            }
        }

        #endregion

        private const uint FlankingMoguId = 60847;
        private const uint PinningArrowId = 60958;

        private const uint MengTheDementedId = 60708;
        private const uint QiangTheMerciless = 60709;
        private const uint SubetaiTheSwift = 60710;
        private const uint ZianOfTheEndlessShadow = 60701;


        private const uint EmperorsRageId = 60396;
        private const uint EmperorsCourageId = 60398;

        readonly uint[] _emperorsStrengthIds = new uint[] { 60397, 60768 };

        private const uint UndyingShadowsId = 60731;
        private const uint EmpyrealFocusId = 60776; // kill these in stage 3.
        private const uint CelestialProtectorId = 60793;
        private const uint EnergyChargeId = 60913; // kill these before they leave the Energy Vortext.
        private const uint ElegonId = 60410;
        private const uint QinxiId = 60399;
        private const uint JanxiId = 60400;

        private readonly uint[] _spiritKingIds = new[] { MengTheDementedId, QiangTheMerciless, SubetaiTheSwift, ZianOfTheEndlessShadow };
        private readonly int[] _volleyIds = new[] { 118094, 118105, 118106 };

        private static LocalPlayer Me
        {
            get { return StyxWoW.Me; }
        }

        [EncounterHandler(0)]
        public Composite RootEncounter()
        {
            return new PrioritySelector();
        }

        [LocationHandler(4227.072, 1595.776, 438.8339, 40, "The Spirit Kings")]
        public Composite TheSpiritKingsEncounter()
        {
            var roomCenterLoc = new WoWPoint(4227.072, 1595.776, 438.8339);
            const int annihilateId = 117948;
            var leftDoorSideloc = new WoWPoint(4190.638, 1580.247, 438.804);
            var rightDoorSideloc = new WoWPoint(4195.618, 1568.984, 441.4225);
            var insideGateLoc = new WoWPoint(4197.687, 1576.985, 438.8339);
            WoWUnit farLeftFlankingMogu = null;
            // dodge flank.
            AddAvoidObject(
                ctx => !Me.IsCasting && farLeftFlankingMogu != null && farLeftFlankingMogu.IsValid && farLeftFlankingMogu.Distance <= 20,
                () => roomCenterLoc,
                31,
                14.25f,
                o => o == farLeftFlankingMogu,
                o =>
                {
                    var mogu = (WoWUnit)o;
                    var rotation = WoWMathHelper.NormalizeRadian(mogu.Rotation);
                    // rotate 90 degrees towards center of flank.
                    var rotationTowardsCenter = WoWMathHelper.NormalizeRadian(rotation + (float)Math.PI / 2);
                    var flankCenterLoc = mogu.Location.RayCast(rotationTowardsCenter, 11.25f);
                    return Me.Location.GetNearestPointOnSegment(flankCenterLoc, flankCenterLoc.RayCast(rotation, 25));
                });
            // todo; remove this after testing.
            AddAvoidObject(ctx => !Me.IsCasting, 3f, FlankingMoguId);

            AddAvoidObject(ctx => !Me.IsCasting, () => roomCenterLoc, 31, 6, UndyingShadowsId);

            WoWUnit activeBoss = null;

            return new PrioritySelector(
                new Decorator(
                    ctx => Me.Combat,
                    new PrioritySelector(
                        ctx =>
                        {
                            farLeftFlankingMogu = GetFarLeftFlankingMogu();
                            return activeBoss = ObjectManager.GetObjectsOfType<WoWUnit>().FirstOrDefault(u => _spiritKingIds.Contains(u.Entry) && u.Attackable);
                        },
                // move inside the room to avoid getting locked out when encounter starts.
                        new Decorator(
                            ctx =>
                            {
                                var tank = ScriptHelpers.Tank;
                                if (Me.Combat || tank == null) return false;
                                return tank.Z > 430 && tank.Location.IsPointLeftOfLine(leftDoorSideloc, rightDoorSideloc) &&
                                       !Me.Location.IsPointLeftOfLine(leftDoorSideloc, rightDoorSideloc);
                            },
                            new Action(ctx => Navigator.MoveTo(insideGateLoc))),
                        new Decorator(
                            ctx => activeBoss != null && activeBoss.Entry == QiangTheMerciless,
                            new PrioritySelector(
                // stack on tank when Qiang is active to split damge from Massive Attacks
                                new Decorator(
                                    ctx =>
                                    activeBoss.CastingSpellId != annihilateId && activeBoss.CurrentTargetGuid > 0 && activeBoss.CurrentTarget.Distance > 4 &&
                                    !AvoidanceManager.IsRunningOutOfAvoid,
                                    new Action(ctx => Navigator.MoveTo(activeBoss.CurrentTarget.Location))),
                // avoid getting hit by annihilate 
                                ScriptHelpers.CreateAvoidUnitAnglesBehavior(
                                    ctx => activeBoss.CastingSpellId == annihilateId && activeBoss.Distance <= 20, () => activeBoss, new ScriptHelpers.AngleSpan(0, 180)))),
                // Avoid getting hit by volley.
                        new Decorator(
                            ctx => activeBoss != null && activeBoss.Entry == SubetaiTheSwift,
                            ScriptHelpers.CreateAvoidUnitAnglesBehavior(
                                ctx => _volleyIds.Contains(activeBoss.CastingSpellId), () => activeBoss, new ScriptHelpers.AngleSpan(0, 100))))));
        }

        private WoWUnit GetFarLeftFlankingMogu()
        {
            WoWUnit mogu = null;
            var flankingMogus = ObjectManager.GetObjectsOfType<WoWUnit>().Where(u => u.Entry == FlankingMoguId && u.HasAura("Overhand Strike")).ToArray();
            if (flankingMogus.Any())
            {
                var rotation = WoWMathHelper.NormalizeRadian(flankingMogus[0].Rotation);
                // rotate 270 degrees and normalize again.
                rotation = WoWMathHelper.NormalizeRadian(rotation + (float)(Math.PI + Math.PI / 2));
                // generate a point to the left side and return the mogu nearest to point.
                var leftPoint = flankingMogus[0].Location.RayCast(rotation, 23);
                mogu = flankingMogus.OrderBy(u => u.Location.DistanceSqr(leftPoint)).FirstOrDefault();
            }
            return mogu;
        }

        [EncounterHandler(60410, "Elegon")]
        public Composite ElegonEncounter()
        {
            // 2nd boss
            const int EnergyCascadeId = 122199; // Avoid this missile, does 5 yard aoe damge at impact location.
            const uint EmpyrealFocusId = 60776; // kill these in stage 3.

            var roomCenterLoc = new WoWPoint(4023.042, 1907.786, 358.7884);

            AddAvoidLocation(ctx => !Me.IsCasting, 8, o => ((WoWMissile)o).ImpactPosition, () => WoWMissile.InFlightMissiles.Where(m => m.SpellId == EnergyCascadeId));
            AddAvoidObject(
                ctx => true,
                2,
                o => o.Entry == EmpyrealFocusId && o.ToUnit().CanSelect,
                o =>
                {
                    var loc = o.Location;
                    return Me.Location.GetNearestPointOnSegment(loc, loc.RayCast(WoWMathHelper.NormalizeRadian(o.Rotation), 15));
                });
            WoWUnit boss = null;
            return new PrioritySelector(
                ctx => boss = ctx as WoWUnit,
                // move off the platform when in phase 3.
                new Decorator(
                    ctx => boss.HasAura("Restart Engine") && _elegonRoomCenterLoc.Distance(Me.Location) < ElegonInnerCircleRadius,
                    new Action(
                        ctx => Navigator.MoveTo(GetElegonInnerCircleMovetoPoints(ElegonInnerCircleRadius + 2).OrderBy(l => l.DistanceSqr(Me.Location)).FirstOrDefault()))),
                // move on the platform if in phase 1 and 2.
                new Decorator(
                    ctx => !boss.HasAura("Restart Engine") && _elegonRoomCenterLoc.Distance(Me.Location) > ElegonInnerCircleRadius - 5,
                    new Action(ctx => Navigator.MoveTo(_elegonRoomCenterLoc))),
                // stay behind boss during phase 1 to aboid getting hit by Celestial Breath.
                new Decorator(
                    ctx => !boss.HasAura("Restart Engine") && boss.IsSafelyFacing(Me, 180),
                    new Action(ctx => Navigator.MoveTo(WoWMathHelper.CalculatePointBehind(boss.Location, WoWMathHelper.NormalizeRadian(boss.Rotation), 6)))),
                new Decorator(ctx => !Me.IsMoving && Me.CurrentTargetGuid > 0 && !Me.IsSafelyFacing(Me.CurrentTarget),
                    new Action(ctx => Me.CurrentTarget.Face())),
                new Decorator(ctx => !Me.IsHealer() && Targeting.Instance.IsEmpty(),
                    new ActionAlwaysSucceed())
                    );
        }

        // 3rd boss 
        //new WoWPoint (3874.398, 1549.394, 362.1991); // radius 50.
        [LocationHandler(3874.398, 1549.394, 362.1991, 100, "Will Of The Emperor")]
        public Composite WillOfTheEmperorEncounter()
        {
            var rangeGroupCenter = WoWPoint.Zero;
            var roomCenterLoc = new WoWPoint(3874.398, 1549.394, 362.1991);

            AddAvoidObject(ctx => !Me.IsCasting, () => roomCenterLoc, 50, 12, u => u.Entry == EmperorsRageId && u.ToUnit().CurrentTargetGuid == Me.Guid);

            WoWUnit boss = null;
            return new PrioritySelector(
                ctx => boss = ctx as WoWUnit,
                new Decorator(
                    ctx =>
                    {
                        if (!Me.HasAura("Titan Gas") || !Me.IsRange())
                            return false;
                        rangeGroupCenter = ScriptHelpers.GetGroupCenterLocation(m => m.IsRange, 15);
                        return rangeGroupCenter.Distance(Me.Location) > 5;
                    },
                    new Action(ctx => Navigator.MoveTo(rangeGroupCenter))));
        }
    }
}